print("import: 'docker'")
import docker

print("import: 'docker.api'")
import docker.api

print("import: 'docker.auth'")
import docker.auth

print("import: 'docker.tls'")
import docker.tls

print("import: 'docker.utils'")
import docker.utils

print("import: 'docker.utils.ports'")
import docker.utils.ports

print("import: 'docker.transport.unixconn'")
import docker.transport.unixconn

